import * as ts from 'typescript';
import { createFilter } from 'rollup-pluginutils';
import { statSync } from 'fs';
import assign from 'object-assign';
import compareVersions from 'compare-versions';
import * as tippex from 'tippex';

var helpersTemplate = "export function __extends(d, b) {\n    for (var p in b) if (b.hasOwnProperty(p)) d[p] = b[p];\n    function __() { this.constructor = d; }\n    d.prototype = b === null ? Object.create(b) : (__.prototype = b.prototype, new __());\n}\n\nexport function __decorate(decorators, target, key, desc) {\n    var c = arguments.length, r = c < 3 ? target : desc === null ? desc = Object.getOwnPropertyDescriptor(target, key) : desc, d;\n    if (typeof Reflect === \"object\" && typeof Reflect.decorate === \"function\") r = Reflect.decorate(decorators, target, key, desc);\n    else for (var i = decorators.length - 1; i >= 0; i--) if (d = decorators[i]) r = (c < 3 ? d(r) : c > 3 ? d(target, key, r) : d(target, key)) || r;\n    return c > 3 && r && Object.defineProperty(target, key, r), r;\n}\n\nexport function __metadata(k, v) {\n    if (typeof Reflect === \"object\" && typeof Reflect.metadata === \"function\") return Reflect.metadata(k, v);\n}\n\nexport function __param(paramIndex, decorator) {\n    return function (target, key) { decorator(target, key, paramIndex); }\n}\n\nexport function __awaiter(thisArg, _arguments, P, generator) {\n    return new (P || (P = Promise))(function (resolve, reject) {\n        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }\n        function rejected(value) { try { step(generator.throw(value)); } catch (e) { reject(e); } }\n        function step(result) { result.done ? resolve(result.value) : new P(function (resolve) { resolve(result.value); }).then(fulfilled, rejected); }\n        step((generator = generator.apply(thisArg, _arguments)).next());\n    });\n}\n";

function endsWith(str, tail) {
    return !tail.length || str.slice(-tail.length) === tail;
}

// Hack around TypeScript's broken handling of `export class` with
// ES6 modules and ES5 script target.
//
// It works because TypeScript transforms
//
//     export class A {}
//
// into something like CommonJS, when we wanted ES6 modules.
//
//     var A = (function () {
//         function A() {
//         }
//         return A;
//     }());
//     exports.A = A;
//
// But
//
//     class A {}
//     export { A };
//
// is transformed into this beauty.
//
//     var A = (function () {
//         function A() {
//         }
//         return A;
//     }());
//     export { A };
//
// The solution is to replace the previous export syntax with the latter.
function fix(code, id) {
    // Erase comments, strings etc. to avoid erroneous matches for the Regex.
    var cleanCode = tippex.erase(code);
    var re = /export\s+(default\s+)?((?:abstract\s+)?class)(?:\s+(\w+))?/g;
    var match;
    while (match = re.exec(cleanCode)) {
        // To keep source maps intact, replace non-whitespace characters with spaces.
        code = erase(code, match.index, match[0].indexOf(match[2]));
        var name = match[3];
        if (match[1]) {
            // TODO: support this too 
            if (!name)
                throw new Error("TypeScript Plugin: cannot export an un-named class (module " + id + ")");
            // Export the name ` as default`.
            name += ' as default';
        }
        // To keep source maps intact, append the injected exports last.
        code += "\nexport { " + name + " };";
    }
    return code;
}
function erase(code, start, length) {
    var end = start + length;
    return code.slice(0, start) +
        code.slice(start, end).replace(/[^\s]/g, ' ') +
        code.slice(end);
}

var jsxOptions = {
    'preserve': ts.JsxEmit.Preserve,
    'react': ts.JsxEmit.React
};
var resolveHost = {
    fileExists: function (filePath) {
        try {
            return statSync(filePath).isFile();
        }
        catch (err) {
            return false;
        }
    }
};
function goodErrors(diagnostic) {
    // All errors except `Cannot compile modules into 'es6' when targeting 'ES5' or lower.`
    return diagnostic.code !== 1204;
}
function typescript(options) {
    options = assign({}, options || {});
    var filter = createFilter(options.include || ['*.ts+(|x)', '**/*.ts+(|x)'], options.exclude || ['*.d.ts', '**/*.d.ts']);
    delete options.include;
    delete options.exclude;
    // Allow users to override the TypeScript version used for transpilation.
    var typescript = options.typescript || ts;
    if (typeof options.jsx === 'string') {
        options.jsx = jsxOptions[options.jsx] || ts.JsxEmit.None;
    }
    options = assign({
        noEmitHelpers: true,
        target: ts.ScriptTarget.ES5,
        module: ts.ModuleKind.ES6,
        sourceMap: true
    }, options);
    return {
        load: function (id) {
            if (id === 'typescript-helpers') {
                return helpersTemplate;
            }
        },
        resolveId: function (importee, importer) {
            // Handle the special `typescript-helpers` import itself.
            if (importee === 'typescript-helpers')
                return 'typescript-helpers';
            if (!importer)
                return null;
            var result;
            if (compareVersions(typescript.version, '1.8.0') < 0) {
                result = typescript.nodeModuleNameResolver(importee, importer, resolveHost);
            }
            else {
                result = typescript.nodeModuleNameResolver(importee, importer, options, resolveHost);
            }
            if (result.resolvedModule && result.resolvedModule.resolvedFileName) {
                if (endsWith(result.resolvedModule.resolvedFileName, '.d.ts')) {
                    return null;
                }
                return result.resolvedModule.resolvedFileName;
            }
            return null;
        },
        transform: function (code, id) {
            if (!filter(id))
                return null;
            var transformed = typescript.transpileModule(fix(code, id), {
                fileName: id,
                reportDiagnostics: true,
                compilerOptions: options
            });
            var diagnostics = transformed.diagnostics.filter(goodErrors);
            var fatalError = false;
            diagnostics.forEach(function (diagnostic) {
                var message = typescript.flattenDiagnosticMessageText(diagnostic.messageText, '\n');
                if (diagnostic.file) {
                    var _a = diagnostic.file.getLineAndCharacterOfPosition(diagnostic.start), line = _a.line, character = _a.character;
                    console.error(diagnostic.file.fileName + "(" + (line + 1) + "," + (character + 1) + "): error TS" + diagnostic.code + ": " + message);
                }
                else {
                    console.error("Error: " + message);
                }
                if (diagnostic.category === ts.DiagnosticCategory.Error) {
                    fatalError = true;
                }
            });
            if (fatalError) {
                throw new Error("There were TypeScript errors transpiling \"" + id + "\"");
            }
            return {
                // Always append an import for the helpers.
                code: transformed.outputText +
                    "\nimport { __extends, __decorate, __metadata, __param, __awaiter } from 'typescript-helpers';",
                // Rollup expects `map` to be an object so we must parse the string
                map: JSON.parse(transformed.sourceMapText)
            };
        }
    };
}

export default typescript;